// AcTree.cpp: implementation of the AcTree class.
//
//////////////////////////////////////////////////////////////////////

#include "AcTree.h"
#include "AcTreeNode.h"
#include "CompoundComponent.h"
#include "Writer.h"
#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

AcTree::AcTree(long numComponents)
{
	m_numComponents = numComponents;
	m_root = new AcTreeNode(m_numComponents);
	assert(m_root != NULL);
	m_root->m_depth = 0;
	for (long i = 0; i < maxAcTreeDepth; i++)
	{
		m_level[i] = NULL;
	}
	m_level[0] = m_root;
	m_numNodes = 0;
	m_numSequences = 0;
}

AcTree::~AcTree()
{
	for (long i = 0; i < maxAcTreeDepth; i++)
	{
		AcTreeNode * node = m_level[i];
		while (node != NULL)
		{
			AcTreeNode * nd = node;
			node = node->m_next;
			delete nd;
		}
	}
}

bool AcTree::checkComponent(CompoundComponent *com)
{
	// this function should be used when the tree is not yet complete
	// to check if this component or its part isn't already in the tree
	// returns true if the tree doesn't contain part of the component

	if (com->getNumComponents() == 0)
	{
		// declare empty component always in the tree
		return false;
	}

	long i;
	for (i = 0; i < com->getNumComponents(); i++)
	{
		AcTreeNode * node = m_root;
		long j = i;

		do 
		{
			node = node->m_ptr[com->getComponentId(j)];
			if (node == NULL)
			{
				// no match, let's try another one
				break;
			}
			else if (node->m_final)
			{
				// final node reached
				return false;
			}
			j++;
		} while (j < com->getNumComponents());
	}
	return true;
}

void AcTree::addComponent(CompoundComponent *com)
{
	// adds a component to the tree
	// only if it or its part isn't already in the tree
	// and if it isn't too big

	if ((com->getNumComponents() <= maxAcTreeDepth) && checkComponent(com))
	{
		/*
		Writer::print("ExcSeq: ");
		com->print();
		Writer::print("\n");
		*/
		AcTreeNode * node = m_root;
		long i = 0;

		while (i < com->getNumComponents())
		{
			AcTreeNode * nd = node->m_ptr[com->getComponentId(i)];
			if (nd == NULL)
			{
				++m_numNodes; 
				nd = new AcTreeNode(m_numComponents);
				assert(nd != NULL);
				// set node depth
				nd->m_depth = node->m_depth + 1;
				// put the node to the appropriate level
				nd->m_next = m_level[i+1];
				m_level[i+1] = nd;
				// set the node's backout to the previous node
				nd->m_backout = node;
				// and put the node to the appropriate tree place
				node->m_ptr[com->getComponentId(i)] = nd;
			} 
			else if (nd->m_final)
			{
				// this should not happen
				// but this is not a time-critical code
				// so let's be prepared
				return;
			}

			// go into the node
			node = nd;
			// and increase the counter (and tree level
			i++;
		}

		// mark the current node as final
		node->m_final = true;
		++m_numSequences;
	}
}

void AcTree::finish()
{
	// this function finishes the tree
	// i.e. makes extended Aho/Corasick tree from it

	// the structure now contains a pure tree and we can rely on it
	// let's calculate the tree depth
	// and point all backout pointers to level up

	// root's backout should be NULL

	AcTreeNode * node = m_root;
	AcTreeNode * nd = NULL;

	long i, j;

	// root's (level 0) backout is NULL - OK
	// level 1's backouts are set to root - OK

	// let's set backouts for level 2 and above

	for (i = 2; i < maxAcTreeDepth; i++)
	{
		node = m_level[i];
		while (node != NULL)
		{
			// in the current node, backout still leads one level up
			nd = node->m_backout;

			// find which symbol leads to the current node
			j = 0;
			while (nd->m_ptr[j] != node)
			{
				++j;
			}

			// j is the 'symbol' that leads to this node

			// in nd's node, backout is a real backout
			nd = nd->m_backout;

			while ((nd != NULL) && (nd->m_ptr[j] == NULL))
			{
				nd = nd->m_backout;
			}

			if (nd == NULL)
			{
				node->m_backout = m_root;
			}
			else
			{
				node->m_backout = nd->m_ptr[j];
			}

			if (node->m_backout->m_final)
			{
				node->m_final = true;
			}

			node = node->m_next;
		}
	}

	// finish the net
	// make transitions for each value in each state

	for (i = maxAcTreeDepth - 1; i >= 0; --i)
	{
		node = m_level[i];
		while (node != NULL)
		{
			for (j = 0; j < m_numComponents; j++)
			{
				if (node->m_ptr[j] == NULL)
				{
					nd = node;
					while ((nd != NULL) && (nd->m_ptr[j] == NULL))
					{
						nd = nd->m_backout;
					}
					if (nd == NULL)
					{
						nd = m_root;
					}
					else
					{
						nd = nd->m_ptr[j];
					}
					node->m_ptr[j] = nd;
				}
			}
			node = node->m_next;
		}
	}

	// structure built
	// now let's change all references to m_final to NULLs

	for (i = 0; i < maxAcTreeDepth; i++)
	{
		node = m_level[i];
		while (node != NULL)
		{
			for (j = 0; j < m_numComponents; j++)
			{
				if (node->m_ptr[j]->m_final)
				{
					node->m_ptr[j] = NULL;
				}
			}
			node = node->m_next;
		}
	}

	// it's done!!!
}

long AcTree::getNumNodes()
{	
	return m_numNodes;
}
